import logging
from typing import Iterable, Optional, Tuple

import httpx
from httpx import AsyncByteStream
from multimethod import multimethod

from httpx_caching import AsyncDictCache, _policy as protocol
from httpx_caching._heuristics import BaseHeuristic
from httpx_caching._models import Response
from httpx_caching._policy import CachingPolicy, Source
from httpx_caching._types import RawHeaders, RawURL
from httpx_caching._utils import ByteStreamWrapper, request_to_raw


class AsyncCachingTransport(httpx.AsyncBaseTransport):
    invalidating_methods = {"PUT", "PATCH", "DELETE"}

    def __init__(
        self,
        transport: httpx.AsyncBaseTransport,
        cache: AsyncDictCache = None,
        cache_etags: bool = True,
        heuristic: BaseHeuristic = None,
        cacheable_methods: Iterable[str] = ("GET",),
        cacheable_status_codes: Iterable[int] = (
            200,
            203,
            300,
            301,
            308,
        ),
    ):
        self.transport = transport

        self.cache = AsyncDictCache() if cache is None else cache
        self.heuristic = heuristic
        self.cacheable_methods = cacheable_methods
        self.cacheable_status_codes = cacheable_status_codes
        self.cache_etags = cache_etags

    async def handle_async_request(
        self,
        method: bytes,
        url: RawURL,
        headers: RawHeaders,
        stream: AsyncByteStream,
        extensions: dict,
    ) -> Tuple[int, RawHeaders, AsyncByteStream, dict]:

        request = httpx.Request(
            method=method,
            url=url,
            headers=headers,
            stream=stream,
        )

        caching_protocol = CachingPolicy(
            request=request,
            cache_etags=self.cache_etags,
            heuristic=self.heuristic,
            cacheable_methods=self.cacheable_methods,
            cacheable_status_codes=self.cacheable_status_codes,
        )

        response, source = await caching_protocol.arun(self.aio_handler)

        response.extensions["from_cache"] = source == Source.CACHE
        return response.to_raw()

    @multimethod
    async def aio_handler(self, action):
        raise NotImplementedError(f"Cannot handle {action}")

    @aio_handler.register
    async def _io_cache_get(
        self, action: protocol.CacheGet
    ) -> Tuple[Optional[Response], Optional[dict]]:
        return await self.cache.aget(action.key)

    @aio_handler.register
    async def _io_cache_delete(self, action: protocol.CacheDelete) -> None:
        await self.cache.adelete(action.key)
        return None

    @aio_handler.register
    async def _io_cache_set(self, action: protocol.CacheSet) -> Optional[Response]:
        if action.deferred:
            # This is a response with a body, so we need to wait for it to be read before we can cache it
            return self.wrap_response_stream(
                action.key, action.response, action.vary_header_values
            )
        else:
            stream = action.response.stream
            # TODO: Are we needlessly recaching the body here? Is this just a header change?
            await self.cache.aset(
                action.key,
                action.response,
                action.vary_header_values,
                b"".join(stream),  # type: ignore
            )
        return None

    @aio_handler.register
    async def _io_make_request(self, action: protocol.MakeRequest) -> Response:
        args = request_to_raw(action.request)
        raw_response = await self.transport.handle_async_request(*args)  # type: ignore
        return Response.from_raw(raw_response)

    @aio_handler.register
    async def _io_close_response_stream(
        self, action: protocol.CloseResponseStream
    ) -> None:
        async for _chunk in action.response.stream:  # type: ignore
            pass
        await action.response.stream.aclose()
        return None

    def wrap_response_stream(
        self, key: str, response: Response, vary_header_values: dict
    ) -> Response:
        response_stream: AsyncByteStream = response.stream  # type: ignore
        wrapped_stream = ByteStreamWrapper(response_stream)
        response.stream = wrapped_stream

        async def callback(response_body: bytes):
            logging.debug(f"saving to cache: {key}")
            await self.cache.aset(key, response, vary_header_values, response_body)

        response.stream.callback = callback
        return response

    async def aclose(self) -> None:
        await self.cache.aclose()
        await self.transport.aclose()  # type: ignore



# SPDX-FileCopyrightText: 2015 Eric Larson
#
# SPDX-License-Identifier: Apache-2.0
from typing import Optional, Tuple

import msgpack
from httpx import ByteStream

from ._models import Response


class Serializer(object):
    def dumps(self, response: Response, vary_header_data: dict, response_body: bytes):
        extensions = response.extensions.copy()
        extensions.pop("real_request", None)
        extensions.pop("close", None)
        extensions.pop("aclose", None)

        data = {
            "response": {
                "body": response_body,
                "headers": response.headers.raw,
                "status_code": response.status_code,
                # TODO: Make sure we don't explode if there's something naughty in extensions
                "extensions": extensions,
            },
            "vary": vary_header_data,
        }

        return b",".join([b"cc=0", msgpack.dumps(data, use_bin_type=True)])

    def loads(self, data: bytes) -> Tuple[Optional[Response], Optional[dict]]:
        # Short circuit if we've been given an empty set of data
        if not data:
            return None, None

        # Determine what version of the serializer the data was serialized
        # with
        try:
            ver, data = data.split(b",", 1)
        except ValueError:
            ver = b"cc=0"

        # Make sure that our "ver" is actually a version and isn't a false
        # positive from a , being in the data stream.
        if ver[:3] != b"cc=":
            data = ver + data
            ver = b"cc=0"

        # Get the version number out of the cc=N
        version = ver.split(b"=", 1)[-1].decode("ascii")

        # Dispatch to the actual load method for the given version
        try:
            return getattr(self, "_loads_v{}".format(version))(data)

        except AttributeError:
            # This is a version we don't have a loads function for, so we'll
            # just treat it as a miss and return None
            return None, None

    def prepare_response(self, cached_data: dict):
        """Construct a response from cached data"""

        cached_response = cached_data["response"]

        status_code = cached_response["status_code"]
        headers = cached_response["headers"]
        stream = ByteStream(cached_response["body"])
        extensions = cached_response["extensions"]

        response = Response.from_raw((status_code, headers, stream, extensions))

        if response.headers.get("transfer-encoding") == "chunked":
            response.headers.pop("transfer-encoding")

        return response, cached_data["vary"]

    def _loads_v0(self, data):
        try:
            cached = msgpack.loads(data, raw=False)
        except ValueError:
            return

        return self.prepare_response(cached)



import dataclasses
from typing import Union

from httpx import ByteStream, Headers

from ._utils import ByteStreamWrapper


@dataclasses.dataclass
class Response:
    """
    Simple wrapper for the raw response returned by a transport.
    """

    status_code: int
    headers: Headers
    stream: Union[ByteStream, ByteStreamWrapper]
    extensions: dict = dataclasses.field(default_factory=dict)

    @classmethod
    def from_raw(cls, raw_response):
        values = list(raw_response)
        values[1] = Headers(values[1])
        if isinstance(values[2], bytes):
            values[2] = [values[2]]
        return cls(*values)

    def to_raw(self):
        raw = []
        for field in dataclasses.fields(self):
            value = getattr(self, field.name)
            if field.name == "headers":
                value = value.raw
            raw.append(value)
        return tuple(raw)



import httpx

from httpx_caching._async._transport import AsyncCachingTransport
from httpx_caching._sync._transport import SyncCachingTransport
from httpx_caching._types import AnyClient


def CachingClient(client: AnyClient, *args, **kwargs) -> AnyClient:
    # TODO: Change this once HTTPX adds mounting functionality
    current_transport = client._transport
    if "transport" not in kwargs:
        kwargs["transport"] = current_transport

    is_async = isinstance(current_transport, httpx.AsyncBaseTransport)
    client._transport = (AsyncCachingTransport if is_async else SyncCachingTransport)(
        *args, **kwargs
    )

    return client



import calendar
import logging
import time
import typing
from copy import copy
from dataclasses import dataclass
from email.utils import parsedate_tz
from enum import Enum
from typing import Awaitable, Callable, Generator, Iterable, Optional, Tuple, Union

from httpx import ByteStream, Headers, Request, codes

from ._heuristics import BaseHeuristic
from ._models import Response
from ._utils import async_callback_generator, sync_callback_generator

logger = logging.getLogger(__name__)

PERMANENT_REDIRECT_STATUSES = (
    301,
    308,
)

INVALIDATING_METHODS = ("PUT", "PATCH", "DELETE")

Source = Enum("Source", ["CACHE", "SERVER"])
Evaluation = Enum("Evaluation", ["GOOD", "INCONCLUSIVE"])
CacheVerb = Enum("CacheVerb", ["GET", "SET", "DELETE"])
VaryData = dict

# Cache actions


@dataclass
class CacheGet:
    key: str


@dataclass
class CacheSet:
    key: str
    response: Response
    vary_header_values: dict
    deferred: bool = False


@dataclass
class CacheDelete:
    key: str


CacheAction = Union[CacheGet, CacheSet, CacheDelete]


# HTTP request related IO actions


@dataclass
class MakeRequest:
    request: Request


@dataclass
class CloseResponseStream:
    response: Response


IOAction = Union[CacheAction, MakeRequest, CloseResponseStream]


AsyncIOCallback = Callable[[IOAction], Awaitable[Optional[Response]]]
SyncIOCallback = Callable[[IOAction], Optional[Response]]


@dataclass
class CachingPolicy:
    request: Request
    cache_etags: bool
    heuristic: Optional[BaseHeuristic]
    cacheable_methods: Iterable[str]
    cacheable_status_codes: Iterable[int]

    @typing.no_type_check
    def run(
        self,
        io_callback: SyncIOCallback,
    ) -> Tuple[Response, Source]:
        # TODO: Shouldn't need to make mypy ignore this should I?
        return sync_callback_generator(
            caching_policy,
            io_callback,
            dict(
                request=self.request,
                cache_etags=self.cache_etags,
                heuristic=self.heuristic,
                cacheable_methods=self.cacheable_methods,
                cacheable_status_codes=self.cacheable_status_codes,
            ),
        )

    @typing.no_type_check
    async def arun(
        self,
        io_callback: AsyncIOCallback,
    ) -> Tuple[Response, Source]:
        return await async_callback_generator(
            caching_policy,
            io_callback,
            dict(
                request=self.request,
                cache_etags=self.cache_etags,
                heuristic=self.heuristic,
                cacheable_methods=self.cacheable_methods,
                cacheable_status_codes=self.cacheable_status_codes,
            ),
        )


def caching_policy(
    request: Request,
    cache_etags: bool,
    heuristic: BaseHeuristic,
    cacheable_methods: Tuple[str],
    cacheable_status_codes: Tuple[int],
) -> Generator[IOAction, Response, Tuple[Response, Source]]:
    cached_response, evaluation = yield from try_from_cache_policy(
        request, cacheable_methods
    )
    logging.debug(f"evaluation: {evaluation}")
    if cached_response and evaluation == Evaluation.GOOD:
        return cached_response, Source.CACHE

    response, source = yield from try_from_server_policy(
        request,
        cached_response,
        heuristic,
        cache_etags,
        cacheable_status_codes,
        cacheable_methods,
    )
    return response, source


def try_from_cache_policy(
    request: Request,
    cacheable_methods: Iterable[str],
) -> Generator[
    CacheAction,
    Tuple[Response, VaryData],
    Union[Tuple[Response, Evaluation], Tuple[None, None]],
]:
    """
    yield cache actions
    expects responses in return
    may finally return valid response as StopIteration value
    """
    # Will only yield GET or DELETE CacheActions. Does not write to cache.
    cache_key = get_cache_key(request)

    if request.method not in cacheable_methods:
        return None, None

    cc = parse_cache_control_directives(request.headers)

    # Bail out if the request insists on fresh data
    if "no-cache" in cc:
        logger.debug('Request header has "no-cache", cache bypassed')
        return None, None

    if cc.get("max-age") == 0:
        logger.debug('Request header has "max_age" as 0, cache bypassed')
        return None, None

    logger.debug(f'Looking up "{cache_key}" in the cache')
    cached_response: Optional[Response]
    cached_vary_data: dict
    cached_response, cached_vary_data = yield CacheGet(cache_key)
    if cached_response is None:
        logger.debug("No cache entry available")
        return None, None

    if not check_vary_headers(request.headers, cached_vary_data):
        logger.debug("Ignoring cache entry due to vary header mismatch")
        return None, None

    # If we have a cached permanent redirect, return it immediately. We
    # don't need to test our response for other headers b/c it is
    # intrinsically "cacheable" as it is Permanent.
    #
    # See:
    #   https://tools.ietf.org/html/rfc7231#section-6.4.2
    #
    # Client can try to refresh the value by repeating the request
    # with cache busting headers as usual (ie no-cache).
    if cached_response.status_code in PERMANENT_REDIRECT_STATUSES:
        msg = (
            "Returning cached permanent redirect response "
            "(ignoring date and etag information)"
        )
        logger.debug(msg)
        return cached_response, Evaluation.GOOD

    if "date" not in cached_response.headers:
        if "etag" not in cached_response.headers:
            # Without date or etag, the cached response can never be used
            # and should be deleted.
            logger.debug("Purging cached response: no date or etag")
            yield CacheDelete(cache_key)
            return None, None
        logger.debug("Ignoring cached response: no date")
        # TODO: Should this return None? Is the cached response now no longer relevant to this request?
        return cached_response, Evaluation.INCONCLUSIVE

    now = time.time()
    # TODO: parsedate_tz might return None (no date value or malformed)
    date = calendar.timegm(parsedate_tz(cached_response.headers["date"]))  # type: ignore
    current_age = max(0, now - date)
    logger.debug("Current age based on date: %i", current_age)

    resp_cc = parse_cache_control_directives(cached_response.headers)

    # determine freshness
    freshness_lifetime = 0

    # Check the max-age pragma in the cache control header
    if "max-age" in resp_cc:
        freshness_lifetime = resp_cc["max-age"]
        logger.debug("Freshness lifetime from max-age: %i", freshness_lifetime)
    # If there isn't a max-age, check for an expires header
    elif "expires" in cached_response.headers:
        expires = parsedate_tz(cached_response.headers["expires"])
        if expires is not None:
            expire_time = calendar.timegm(expires) - date  # type: ignore
            freshness_lifetime = max(0, expire_time)
            logger.debug("Freshness lifetime from expires: %i", freshness_lifetime)

    # Determine if we are setting freshness limit in the
    # request. Note, this overrides what was in the response.
    if "max-age" in cc:
        freshness_lifetime = cc["max-age"]
        logger.debug("Freshness lifetime from request max-age: %i", freshness_lifetime)

    if "min-fresh" in cc:
        min_fresh = cc["min-fresh"]
        # adjust our current age by our min fresh
        current_age += min_fresh
        logger.debug("Adjusted current age from min-fresh: %i", current_age)

    # Return entry if it is fresh enough
    if freshness_lifetime > current_age:
        logger.debug('The response is "fresh", returning cached response')
        logger.debug("%i > %i", freshness_lifetime, current_age)
        return cached_response, Evaluation.GOOD

    # we're not fresh. If we don't have an Etag, clear it out
    if "etag" not in cached_response.headers:
        logger.debug('The cached response is "stale" with no etag, purging')
        yield CacheDelete(cache_key)
        return None, None

    # No conclusive response yet.
    return cached_response, Evaluation.INCONCLUSIVE


def try_from_server_policy(
    request: Request,
    cached_response: Optional[Response],
    heuristic: BaseHeuristic,
    cache_etags: bool,
    cacheable_status_codes: Iterable[int],
    cacheable_methods: Iterable[str],
) -> Generator[IOAction, Response, Tuple[Response, Source]]:
    cache_key = get_cache_key(request)
    logging.debug(f"we have this from the cache: {cached_response}")
    updated_headers = request.headers.copy()
    if cached_response:
        # Add conditional headers based on cached response
        for source, target in [
            ("etag", "If-None-Match"),
            ("last-modified", "If-Modified-Since"),
        ]:
            if source in cached_response.headers:
                updated_headers[target] = cached_response.headers[source]

    request = Request(
        method=request.method,
        url=request.url,
        headers=updated_headers,
        stream=request.stream,
    )
    server_response = yield MakeRequest(request)

    # See if we should invalidate the cache.
    if is_invalidating_method(request.method) and not codes.is_error(
        server_response.status_code
    ):
        yield CacheDelete(cache_key)

    if request.method not in cacheable_methods:
        return server_response, Source.SERVER

    # Check for any heuristics that might update headers
    # before trying to cache.
    if heuristic:
        # TODO: don't modify things, return things.
        heuristic.apply(server_response.headers, server_response.status_code)

    # apply any expiration heuristics
    if server_response.status_code == 304:
        # Make sure to clean up the ETag response stream just in case.
        # Compliant servers will not return a body with ETag responses
        yield CloseResponseStream(server_response)

        # We must have sent an ETag request. This could mean
        # that we've been expired already or that we simply
        # have an ETag. In either case, we want to try and
        # update the cache if that is the case.
        if cached_response:
            updated_cached_response = update_with_304_response(
                cached_response, new_response_headers=server_response.headers
            )
            vary_header_values = get_vary_headers(
                request.headers, updated_cached_response
            )
            yield CacheSet(cache_key, updated_cached_response, vary_header_values)
            return updated_cached_response, Source.CACHE

        return server_response, Source.SERVER

    # We have a new response, let's make any changes necessary to the cache (store/delete)
    cache_exists = bool(cached_response)
    cache_action = cache_response_action(
        request,
        server_response,
        cache_exists,
        cache_etags,
        cacheable_status_codes,
    )
    if cache_action:
        wrapped_stream_response = yield cache_action
        if wrapped_stream_response:
            server_response = wrapped_stream_response

    return server_response, Source.SERVER


def cache_response_action(
    request: Request,
    server_response: Response,
    cache_exists: bool,
    cache_etags: bool,
    cacheable_status_codes: Iterable[int],
) -> Optional[Union[CacheSet, CacheDelete]]:
    """
    Algorithm for caching responses.

    Does some checks on request and response and deletes cache if appropriate

    Then either:
    No cache
    Cache immediately with no body for redirects
    Cache with body, this must be deferred.

    Returns:
    May return a request that has had its stream wrapped to trigger caching once read.
    """
    cache_key = get_cache_key(request)

    # From httplib2: Don't cache 206's since we aren't going to
    #                handle byte range requests
    if server_response.status_code not in cacheable_status_codes:
        logger.debug(
            "Status code %s not in %s",
            server_response.status_code,
            cacheable_status_codes,
        )
        return None

    logger.debug('Updating cache with response from "%s"', cache_key)

    # TODO: Do this once on the request/response?
    cc_req = parse_cache_control_directives(request.headers)
    cc = parse_cache_control_directives(server_response.headers)

    # Delete it from the cache if we happen to have it stored there
    no_store = False
    if "no-store" in cc:
        no_store = True
        logger.debug('Response header has "no-store"')
    if "no-store" in cc_req:
        no_store = True
        logger.debug('Request header has "no-store"')
    if no_store and cache_exists:
        logger.debug('Purging existing cache entry to honor "no-store"')
        return CacheDelete(cache_key)
    if no_store:
        return None

    # https://tools.ietf.org/html/rfc7234#section-4.1:
    # A Vary header field-value of "*" always fails to match.
    # Storing such a response leads to a deserialization warning
    # during cache lookup and is not allowed to ever be served,
    # so storing it can be avoided.
    if "*" in server_response.headers.get("vary", ""):
        logger.debug('Response header has "Vary: *"')
        return None

    # If we've been given an etag, then keep the response
    if cache_etags and "etag" in server_response.headers:
        logger.debug("Caching due to etag")

    # Add to the cache any permanent redirects. We do this before looking
    # that the Date headers.
    elif int(server_response.status_code) in PERMANENT_REDIRECT_STATUSES:
        logger.debug("Caching permanent redirect")
        response_body = b""
        response = Response(
            server_response.status_code,
            server_response.headers,
            # TODO: This is naff, maybe we just use httpx.Response
            ByteStream(response_body),
        )
        vary_header_values = get_vary_headers(request.headers, response)
        return CacheSet(cache_key, response, vary_header_values)

    # Add to the cache if the response headers demand it. If there
    # is no date header then we can't do anything about expiring
    # the cache.
    elif "date" in server_response.headers:
        # cache when there is a max-age > 0
        if "max-age" in cc and cc["max-age"] > 0:
            logger.debug("Caching b/c date exists and max-age > 0")

        # If the request can expire, it means we should cache it
        # in the meantime.
        elif "expires" in server_response.headers:
            if server_response.headers["expires"]:
                logger.debug("Caching b/c of expires header")
        else:
            return None
    else:
        return None

    vary_header_values = get_vary_headers(request.headers, server_response)
    return CacheSet(cache_key, server_response, vary_header_values, deferred=True)


def get_cache_key(request: Request) -> str:
    return str(request.url)


def is_invalidating_method(method: str):
    return method in INVALIDATING_METHODS


def parse_cache_control_directives(headers: Headers):
    known_directives = {
        # https://tools.ietf.org/html/rfc7234#section-5.2
        "max-age": (int, True),
        "max-stale": (int, False),
        "min-fresh": (int, True),
        "no-cache": (None, False),
        "no-store": (None, False),
        "no-transform": (None, False),
        "only-if-cached": (None, False),
        "must-revalidate": (None, False),
        "public": (None, False),
        "private": (None, False),
        "proxy-revalidate": (None, False),
        "s-maxage": (int, True),
    }

    cc_headers = headers.get("cache-control", "")

    retval = {}  # type: ignore

    for cc_directive in cc_headers.split(","):
        if not cc_directive.strip():
            continue

        parts = cc_directive.split("=", 1)
        directive = parts[0].strip()

        try:
            typ, required = known_directives[directive]
        except KeyError:
            logger.debug("Ignoring unknown cache-control directive: %s", directive)
            continue

        if not typ or not required:
            retval[directive] = None
        if typ:
            try:
                retval[directive] = typ(parts[1].strip())
            except IndexError:
                if required:
                    logger.debug(
                        "Missing value for cache-control " "directive: %s",
                        directive,
                    )
            except ValueError:
                logger.debug(
                    "Invalid value for cache-control directive " "%s, must be %s",
                    directive,
                    typ.__name__,
                )

    return retval


def update_with_304_response(
    cached_response: Response, new_response_headers: Headers
) -> Response:
    """On a 304 we will get a new set of headers that we want to
    update our cached value with, assuming we have one.

    This should only ever be called when we've sent an ETag and
    gotten a 304 as the response.
    """
    updated_response = copy(cached_response)

    # Lets update our headers with the headers from the new request:
    # http://tools.ietf.org/html/draft-ietf-httpbis-p4-conditional-26#section-4.1
    #
    # The server isn't supposed to send headers that would make
    # the cached body invalid. But... just in case, we'll be sure
    # to strip out ones we know that might be problematic due to
    # typical assumptions.
    excluded_headers = ["content-length"]

    updated_response.headers.update(
        dict(
            (k, v)
            for k, v in new_response_headers.items()
            # TODO: Don't think .lower() is necessary
            if k.lower() not in excluded_headers
        )
    )

    # we want a 200 b/c we have content via the cache
    updated_response.status_code = 200

    return updated_response


def check_vary_headers(request_headers: Headers, cached_vary_data: dict) -> bool:
    """Verify our vary headers match."""
    # Ensure that the Vary headers for the cached response match our
    # request
    # TODO: this should not be here, no reason for request headers to be so deep in deserialization.
    for header, value in cached_vary_data.items():
        if request_headers.get(header, None) != value:
            return False

    return True


def get_vary_headers(request_headers: Headers, response: Response):
    """Get vary headers values for persisting in the cache for later checking"""
    vary = {}

    # Construct our vary headers
    if "vary" in response.headers:
        varied_headers = response.headers["vary"].split(",")
        for header in varied_headers:
            header = header.strip()
            header_value = request_headers.get(header, None)
            vary[header] = header_value

    return vary



# SPDX-FileCopyrightText: 2015 Eric Larson
#
# SPDX-License-Identifier: Apache-2.0

import calendar
import time
from datetime import datetime
from email.utils import formatdate, parsedate
from pprint import pprint

import httpx
import pytest
from httpx import Headers
from mock import Mock

from httpx_caching._heuristics import (
    TIME_FMT,
    BaseHeuristic,
    ExpiresAfterHeuristic,
    LastModifiedHeuristic,
    OneDayCacheHeuristic,
)
from tests.conftest import cache_hit, make_async_client

pytestmark = pytest.mark.asyncio


class TestHeuristicWithoutWarning(object):
    def setup(self):
        class NoopHeuristic(BaseHeuristic):
            warning = Mock()

            def update_headers(self, resp_headers, resp_status):
                return {}

        self.heuristic = NoopHeuristic()
        self.async_client = make_async_client(heuristic=self.heuristic)

    async def test_no_header_change_means_no_warning_header(self, url):
        the_url = url + "optional_cacheable_request"
        await self.async_client.get(the_url)

        assert not self.heuristic.warning.called


class TestHeuristicWith3xxResponse(object):
    def setup(self):
        class DummyHeuristic(BaseHeuristic):
            def update_headers(self, resp_headers, resp_status):
                return {"x-dummy-header": "foobar"}

        self.async_client = make_async_client(heuristic=DummyHeuristic())

    async def test_heuristic_applies_to_301(self, url):
        the_url = url + "permanent_redirect"
        resp = await self.async_client.get(the_url)
        assert "x-dummy-header" in resp.headers

    async def test_heuristic_applies_to_304(self, url):
        the_url = url + "conditional_get"
        resp = await self.async_client.get(the_url)
        assert "x-dummy-header" in resp.headers


class TestOneDayCache(object):
    def setup(self):
        self.async_client = make_async_client(heuristic=OneDayCacheHeuristic())

    async def test_cache_for_one_day(self, url):
        the_url = url + "optional_cacheable_request"
        r = await self.async_client.get(the_url)

        assert "expires" in r.headers
        assert "warning" in r.headers

        pprint(dict(r.headers))

        r = await self.async_client.get(the_url)
        pprint(dict(r.headers))
        assert cache_hit(r)


class TestExpiresAfter(object):
    def setup(self):
        self.async_client = make_async_client(heuristic=ExpiresAfterHeuristic(days=1))

    async def test_expires_after_one_day(self, url):
        the_url = url + "no_cache"
        resp = httpx.get(the_url)
        assert resp.headers["cache-control"] == "no-cache"

        r = await self.async_client.get(the_url)

        assert "expires" in r.headers
        assert "warning" in r.headers
        assert r.headers["cache-control"] == "public"

        r = await self.async_client.get(the_url)
        assert cache_hit(r)


class TestLastModified(object):
    def setup(self):
        self.async_client = make_async_client(heuristic=LastModifiedHeuristic())

    async def test_last_modified(self, url):
        the_url = url + "optional_cacheable_request"
        r = await self.async_client.get(the_url)

        assert "expires" in r.headers
        assert "warning" not in r.headers

        pprint(dict(r.headers))

        r = await self.async_client.get(the_url)
        pprint(dict(r.headers))
        assert cache_hit(r)


def datetime_to_header(dt):
    return formatdate(calendar.timegm(dt.timetuple()))


@pytest.mark.sync
class TestModifiedUnitTests(object):
    def last_modified(self, period):
        return time.strftime(TIME_FMT, time.gmtime(self.time_now - period))

    def setup(self):
        self.heuristic = LastModifiedHeuristic()
        self.time_now = time.time()
        day_in_seconds = 86400
        self.year_ago = self.last_modified(day_in_seconds * 365)
        self.week_ago = self.last_modified(day_in_seconds * 7)
        self.day_ago = self.last_modified(day_in_seconds)
        self.now = self.last_modified(0)

        # NOTE: We pass in a negative to get a positive... Probably
        #       should refactor.
        self.day_ahead = self.last_modified(-day_in_seconds)

    def test_no_expiry_is_inferred_when_no_last_modified_is_present(self):
        assert self.heuristic.update_headers({}, 200) == {}

    def test_expires_is_not_replaced_when_present(self):
        headers = {"Expires": self.day_ahead}
        assert self.heuristic.update_headers(Headers(headers), 200) == {}

    def test_last_modified_is_used(self):
        headers = {"Date": self.now, "Last-Modified": self.week_ago}
        modified = self.heuristic.update_headers(Headers(headers), 200)
        assert ["expires"] == list(modified.keys())
        assert datetime(*parsedate(modified["expires"])[:6]) > datetime.now()  # type: ignore

    def test_last_modified_is_not_used_when_cache_control_present(self):
        headers = {
            "Date": self.now,
            "Last-Modified": self.week_ago,
            "Cache-Control": "private",
        }

        assert self.heuristic.update_headers(Headers(headers), 200) == {}

    def test_last_modified_is_not_used_when_status_is_unknown(self):
        headers = {"Date": self.now, "Last-Modified": self.week_ago}
        status = 299
        assert self.heuristic.update_headers(Headers(headers), status) == {}

    def test_last_modified_is_used_when_cache_control_public(self):
        headers = {
            "Date": self.now,
            "Last-Modified": self.week_ago,
            "Cache-Control": "public",
        }
        modified = self.heuristic.update_headers(Headers(headers), 200)
        assert ["expires"] == list(modified.keys())
        assert datetime(*parsedate(modified["expires"])[:6]) > datetime.now()  # type: ignore

    def test_expiry_is_no_more_than_twenty_four_hours(self):
        headers = {"Date": self.now, "Last-Modified": self.year_ago}
        modified = self.heuristic.update_headers(Headers(headers), 200)
        assert ["expires"] == list(modified.keys())
        assert self.day_ahead == modified["expires"]



# SPDX-FileCopyrightText: 2015 Eric Larson
#
# SPDX-License-Identifier: Apache-2.0

from __future__ import print_function

import pytest
from freezegun import freeze_time

from .conftest import cache_hit, make_client


class TestMaxAge(object):
    @pytest.fixture()
    def client(self, url):
        self.url = url
        client = make_client()

        return client

    def test_client_max_age_0(self, client):
        """
        Making sure when the client uses max-age=0 we don't get a
        cached copy even though we're still fresh.
        """
        cache = client._transport.cache

        print("first request")
        r = client.get(self.url)
        assert cache.get(self.url)

        print("second request")
        r = client.get(self.url, headers={"Cache-Control": "max-age=0"})

        # don't remove from the cache
        assert cache.get(self.url)
        assert not cache_hit(r)

    def test_client_max_age_3600(self, client):
        """
        Verify we get a cached value when the client has a
        reasonable max-age value.
        """
        cache = client._transport.cache
        with freeze_time("2012-01-14"):
            r = client.get(self.url)
            assert cache.get(self.url)

            # request that we don't want a new one unless
            r = client.get(self.url, headers={"Cache-Control": "max-age=3600"})
            assert cache_hit(r)

        # now lets grab one that forces a new request b/c the cache has expired
        r = client.get(self.url)
        assert not cache_hit(r)



# SPDX-FileCopyrightText: 2015 Eric Larson
#
# SPDX-License-Identifier: Apache-2.0

import msgpack

from httpx_caching._models import Response
from httpx_caching._serializer import Serializer


class TestSerializer(object):
    def setup(self):
        self.serializer = Serializer()
        self.response_data = {
            "response": {
                "body": b"Hello World",
                "headers": {
                    "Content-Type": "text/plain",
                    "Expires": "87654",
                    "Cache-Control": "public",
                },
                "status_code": 200,
                "extensions": {},
            },
            "vary": {},
        }

    def test_read_version_v0(self):
        resp, _vary_fields = self.serializer._loads_v0(
            msgpack.dumps(self.response_data)
        )
        assert resp.stream.read() == b"Hello World"

    def test_dumps(self):
        assert self.serializer.dumps(
            Response.from_raw((200, {}, b"foo", {})),
            {},
            b"foo",
        )



from typing import Optional, Tuple

from httpx_caching._models import Response
from httpx_caching._serializer import Serializer
from httpx_caching._utils import AsyncLock


class AsyncDictCache:
    def __init__(self, serializer: Optional[Serializer] = None) -> None:
        self.serializer = serializer if serializer else Serializer()
        self.data: dict = {}
        self.lock = None

    def get_lock(self):
        if not self.lock:
            self.lock = AsyncLock()
        return self.lock

    async def aget(self, key: str) -> Tuple[Optional[Response], Optional[dict]]:
        return self.serializer.loads(self.data.get(key, None))

    async def aset(
        self, key: str, response: Response, vary_header_data: dict, response_body: bytes
    ) -> None:
        async with self.get_lock():
            self.data.update(
                {key: self.serializer.dumps(response, vary_header_data, response_body)}
            )

    async def adelete(self, key: str) -> None:
        async with self.get_lock():
            if key in self.data:
                self.data.pop(key)

    async def aclose(self):
        pass



from httpx_caching._async._cache import AsyncDictCache
from httpx_caching._async._transport import AsyncCachingTransport
from httpx_caching._sync._cache import SyncDictCache
from httpx_caching._sync._transport import SyncCachingTransport
from httpx_caching._wrapper import CachingClient

from ._heuristics import (
    ExpiresAfterHeuristic,
    LastModifiedHeuristic,
    OneDayCacheHeuristic,
)

__version__ = "0.1.0"

__all__ = [
    "AsyncCachingTransport",
    "SyncCachingTransport",
    "CachingClient",
    "AsyncDictCache",
    "SyncDictCache",
    "ExpiresAfterHeuristic",
    "LastModifiedHeuristic",
    "OneDayCacheHeuristic",
]



# SPDX-FileCopyrightText: 2015 Eric Larson
#
# SPDX-License-Identifier: Apache-2.0


import pytest
from freezegun import freeze_time
from httpx import AsyncClient, Limits, Request, Timeout

from httpx_caching import AsyncCachingTransport, AsyncDictCache
from tests.conftest import cache_hit, raw_resp

pytestmark = pytest.mark.asyncio


def get_last_request(client):
    (
        method,
        url,
        headers,
        stream,
        _ext,
    ) = client._transport.transport.handle_async_request.call_args[0]
    return Request(
        method=method,
        url=url,
        headers=headers,
        stream=stream,
    )


@pytest.fixture()
async def async_client(mocker):
    async_client = AsyncClient()
    transport = AsyncCachingTransport(
        transport=async_client._transport,
        cache=AsyncDictCache(),
    )
    async_client._transport = transport

    mocker.patch.object(
        transport.transport,
        "handle_async_request",
        wraps=transport.transport.handle_async_request,
    )

    yield async_client

    await async_client.aclose()


class TestETag(object):
    """Test our equal priority caching with ETags

    Equal Priority Caching is a term I've defined to describe when
    ETags are cached orthgonally from Time Based Caching.
    """

    async def test_etags_get_example(self, async_client, url):
        """RFC 2616 14.26

        The If-None-Match request-header field is used with a method to make
        it conditional. A client that has one or more entities previously
        obtained from the resource can verify that none of those entities
        is current by including a list of their associated entity tags in
        the If-None-Match header field. The purpose of this feature is to
        allow efficient updates of cached information with a minimum amount
        of transaction overhead

        If any of the entity tags match the entity tag of the entity that
        would have been returned in the response to a similar GET request
        (without the If-None-Match header) on that resource, [...] then
        the server MUST NOT perform the requested method, [...]. Instead, if
        the request method was GET or HEAD, the server SHOULD respond with
        a 304 (Not Modified) response, including the cache-related header
        fields (particularly ETag) of one of the entities that matched.

        (Paraphrased) A server may provide an ETag header on a response. On
        subsequent queries, the client may reference the value of this Etag
        header in an If-None-Match header; on receiving such a header, the
        server can check whether the entity at that URL has changed from the
        clients last version, and if not, it can return a 304 to indicate
        the client can use it's current representation.
        """
        r1 = await async_client.get(url + "etag")
        # make sure we cached it
        assert await async_client._transport.cache.aget(url + "etag")

        # make the same request
        r2 = await async_client.get(url + "etag")
        assert cache_hit(r2)
        assert raw_resp(r2) == raw_resp(r1)

        # make the same request a 3rd time to make sure we don't mess anything up
        # after a cache hit
        r3 = await async_client.get(url + "etag")
        assert cache_hit(r3)
        assert raw_resp(r3) == raw_resp(r1)

        # tell the server to change the etags of the response
        await async_client.get(url + "update_etag")

        r4 = await async_client.get(url + "etag")
        assert not cache_hit(r4)
        assert raw_resp(r4) != raw_resp(r1)

        r5 = await async_client.get(url + "etag")
        assert cache_hit(r5)
        assert raw_resp(r5) == raw_resp(r4)


class TestDisabledETags(object):
    """Test our use of ETags when the response is stale and the
    response has an ETag.
    """

    async def test_expired_etags_if_none_match_response(self, async_client, url):
        """Make sure an expired response that contains an ETag uses
        the If-None-Match header.
        """
        # Cache an old etag response
        with freeze_time("2012-01-14"):
            await async_client.get(url + "etag")

        assert await async_client._transport.cache.aget(url + "etag")

        r2 = await async_client.get(url + "etag")
        assert cache_hit(r2)

        real_request = get_last_request(async_client)
        assert "if-none-match" in real_request.headers
        assert r2.status_code == 200


class TestReleaseConnection(object):
    """
    On 304s we still make a request using our connection pool, yet
    we do not call the parent adapter, which releases the connection
    back to the pool. This test ensures that when the parent `get`
    method is not called we consume the response (which should be
    empty according to the HTTP spec) and release the connection.
    """

    @pytest.mark.timeout(5)
    async def test_not_modified_releases_connection(self, url):
        async_client = AsyncClient(
            timeout=Timeout(1, pool=0.1),
            limits=Limits(max_connections=1, max_keepalive_connections=1),
        )
        async_client._transport = AsyncCachingTransport(
            transport=async_client._transport,
        )

        # make sure the pool doesn't time out
        for _i in range(3):
            await async_client.get(url + "etag")

        await async_client.aclose()



# SPDX-FileCopyrightText: 2015 Eric Larson
#
# SPDX-License-Identifier: Apache-2.0
"""
Test for supporting streamed responses (Transfer-Encoding: chunked)
"""
import pytest

from tests.conftest import cache_hit

pytestmark = pytest.mark.asyncio


class TestChunkedResponses(object):
    async def test_cache_chunked_response(self, url, async_client):
        """
        Verify that an otherwise cacheable response is cached when the
        response is chunked.
        """
        url = url + "stream"
        r = await async_client.get(url)
        print(async_client._transport.cache.data)
        assert r.headers.get("transfer-encoding") == "chunked"

        r = await async_client.get(url, headers={"Cache-Control": "max-age=3600"})
        assert cache_hit(r)

    # TODO: Test that stream is only cached the first time it is iterated over
    async def test_stream_is_cached(self, url, async_client):
        async with async_client.stream("GET", url + "stream") as resp_1:
            content_1 = await resp_1.aread()

        async with async_client.stream("GET", url + "stream") as resp_2:
            content_2 = await resp_2.aread()

        assert not cache_hit(resp_1)
        assert cache_hit(resp_2)
        assert content_1 == content_2

    async def test_stream_is_not_cached_when_content_is_not_read(
        self, url, async_client
    ):
        async with async_client.stream("GET", url + "stream"):
            pass
        async with async_client.stream("GET", url + "stream") as resp:
            assert not cache_hit(resp)



from typing import List, Optional, Tuple, TypeVar, Union

from httpcore import AsyncHTTPTransport, SyncHTTPTransport
from httpx import AsyncClient, Client

RawURL = Tuple[bytes, bytes, Optional[int], bytes]
RawHeaders = List[Tuple[bytes, bytes]]

AnyClient = TypeVar("AnyClient", Client, AsyncClient)
AnyTransport = Union[SyncHTTPTransport, AsyncHTTPTransport]



import logging
import threading
from typing import (
    AsyncIterator,
    Awaitable,
    Callable,
    Generator,
    Iterator,
    Optional,
    Tuple,
    TypeVar,
    Union,
)

import anyio
import httpx

AsyncLock = anyio.Lock
SyncLock = threading.Lock


class ByteStreamWrapper:
    def __init__(
        self,
        stream: Union[httpx.SyncByteStream, httpx.AsyncByteStream],
        callback: Optional[Callable] = None,
    ) -> None:
        """
        A wrapper around a stream that calls a callback once with
        the full contents of the stream after it has been fully read.
        """
        self.stream = stream
        self.callback = callback or (lambda *args, **kwargs: None)

        self.buffer = bytearray()
        self.callback_called = False

    def _on_read_finish(self):
        if not self.callback_called:
            self.callback(bytes(self.buffer))
            self.callback_called = True

    async def _a_on_read_finish(self):
        if not self.callback_called:
            await self.callback(bytes(self.buffer))
            self.callback_called = True

    def __iter__(self) -> Iterator[bytes]:
        for chunk in self.stream:  # type: ignore
            self.buffer.extend(chunk)
            yield chunk
        self._on_read_finish()

    async def __aiter__(self) -> AsyncIterator[bytes]:
        async for chunk in self.stream:  # type: ignore
            self.buffer.extend(chunk)
            yield chunk
        await self._a_on_read_finish()

    def close(self) -> None:
        self.stream.close()  # type: ignore

    async def aclose(self) -> None:
        await self.stream.aclose()  # type: ignore


YieldType = TypeVar("YieldType")
SendType = TypeVar("SendType")
ReturnType = TypeVar("ReturnType")


async def async_callback_generator(
    genfunction: Callable[..., Generator[YieldType, SendType, ReturnType]],
    callback: Callable[[YieldType], Awaitable[SendType]],
    kwargs: dict,
):
    gen = genfunction(**kwargs)
    try:
        yielded = next(gen)
        while True:
            logging.debug(f"action: {yielded}")
            to_send = await callback(yielded)
            logging.debug(f"result: {to_send}")
            yielded = gen.send(to_send)
    except StopIteration as e:
        return e.value


def sync_callback_generator(
    genfunction: Callable[..., Generator[YieldType, SendType, ReturnType]],
    callback: Callable[[YieldType], SendType],
    kwargs: dict,
):
    gen = genfunction(**kwargs)
    try:
        yielded = next(gen)
        while True:
            logging.debug(f"action: {yielded}")
            to_send = callback(yielded)
            logging.debug(f"result: {to_send}")
            yielded = gen.send(to_send)
    except StopIteration as e:
        return e.value


def request_to_raw(request: httpx.Request) -> Tuple:
    return (
        request.method.encode("ascii"),
        request.url.raw,
        request.headers.raw,
        request.stream,
        {},  # TODO: need to sort out ext preservation
    )



import os

from setuptools import find_packages, setup

from httpx_caching import __version__

with open(os.path.join(os.getcwd(), "README.md"), encoding="utf-8") as f:
    long_description = f.read()


setup(
    name="http-caching4",
    version=__version__,
    description="Caching support for Async httpx Client. Cloned from https://github.com/kovan/httpx-caching, https://github.com/johtso/httpx-caching",
    long_description=long_description,
    long_description_content_type="text/markdown",
    url="https://github.com/ms280690/httpx-caching4.git",
    author="Mehul Solanki",
    author_email="ms280690@gmail.com",
    packages=find_packages(),
    include_package_data=True,
    install_requires=[
        "httpx==0.18.*",
        "msgpack",
        "anyio",
        "multimethod",
    ],
    extras_require={
        "dev": [
            "pytest",
            "pytest-cov",
            "pytest-asyncio",
            "pytest-mock",
            "pytest-timeout",
            "mock",
            "types-mock",
            "cherrypy",
            "freezegun",
            "types-freezegun",
            "autoflake",
            "black",
            "isort",
            "autoflake",
            "flake8",
            "flake8-bugbear",
            "mypy",
            "unasync",
            "types-dataclasses",
        ],
    },
    test_suite="tests",
    python_requires=">=3.8",
)


# SPDX-FileCopyrightText: 2015 Eric Larson
#
# SPDX-License-Identifier: Apache-2.0

import mock
import pytest

from httpx_caching import AsyncDictCache
from tests.conftest import cache_hit

pytestmark = pytest.mark.asyncio


class TestClientActions(object):
    async def test_get_caches(self, url, async_client):
        await async_client.get(url)
        r2 = await async_client.get(url)
        assert cache_hit(r2)

    async def test_get_with_no_cache_does_not_cache(self, url, async_client):
        await async_client.get(url)
        r2 = await async_client.get(url, headers={"Cache-Control": "no-cache"})
        assert not cache_hit(r2)

    async def test_put_invalidates_cache(self, url, async_client):
        await async_client.get(url)
        await async_client.put(url, data={"foo": "bar"})
        r3 = await async_client.get(url)
        assert not cache_hit(r3)

    async def test_patch_invalidates_cache(self, url, async_client):
        await async_client.get(url)
        await async_client.patch(url, data={"foo": "bar"})
        r3 = await async_client.get(url)
        assert not cache_hit(r3)

    async def test_delete_invalidates_cache(self, url, async_client):
        await async_client.get(url)
        await async_client.delete(url)
        r3 = await async_client.get(url)
        assert not cache_hit(r3)

    @pytest.mark.xfail
    async def test_close(self, url, async_client):
        mock_cache = mock.Mock(spec=AsyncDictCache)
        async_client._transport.cache = mock_cache

        # TODO: httpx does not close transport if nothing has been done with the client
        # await async_client.get(url)

        await async_client.aclose()
        assert mock_cache.close.called



# SPDX-FileCopyrightText: 2015 Eric Larson
#
# SPDX-License-Identifier: Apache-2.0

from urllib.parse import urljoin

import pytest

from httpx_caching import AsyncDictCache
from tests.conftest import cache_hit, make_async_client

pytestmark = pytest.mark.asyncio


class TestVary(object):
    @pytest.fixture()
    def cache(self):
        return AsyncDictCache()

    @pytest.fixture()
    async def async_client(self, cache):
        async_client = make_async_client(cache=cache)
        yield async_client
        await async_client.aclose()

    def assert_cached_equal(self, cached, resp):
        print(cached, resp)
        # remove any transfer-encoding headers as they don't apply to
        # a cached value
        if "chunked" in resp.headers.get("transfer-encoding", ""):
            resp.headers.pop("transfer-encoding")

        assert [cached.stream.read(), cached.headers, cached.status_code] == [
            resp.content,
            resp.headers,
            resp.status_code,
        ]

    async def test_vary_example(self, async_client, cache, url):
        """RFC 2616 13.6

        When the cache receives a subsequent request whose Request-URI
        specifies one or more cache entries including a Vary header field,
        the cache MUST NOT use such a cache entry to construct a response
        to the new request unless all of the selecting request-headers
        present in the new request match the corresponding stored
        request-headers in the original request.

        Or, in simpler terms, when you make a request and the server
        returns defines a Vary header, unless all the headers listed
        in the Vary header are the same, it won't use the cached
        value.
        """
        vary_url = urljoin(url, "/vary_accept")

        r = await async_client.get(vary_url, headers={"foo": "a"})
        cached_response, _vary_data = await cache.aget(vary_url)

        # make the same request
        resp = await async_client.get(vary_url, headers={"foo": "b"})
        assert cache_hit(resp)

        # make a similar request, changing the accept header
        resp = await async_client.get(
            vary_url, headers={"Accept": "text/plain, text/html", "foo": "c"}
        )

        with pytest.raises(AssertionError):
            self.assert_cached_equal(cached_response, resp)
        assert not cache_hit(resp)

        # Just confirming two things here:
        #
        #   1) The server used the vary header
        #   2) We have more than one header we vary on
        #
        # The reason for this is that when we don't specify the header
        # in the request, it is considered the same in terms of
        # whether or not to use the cached value.
        assert "vary" in r.headers
        assert len(r.headers["vary"].replace(" ", "").split(",")) == 2



# SPDX-FileCopyrightText: 2015 Eric Larson
#
# SPDX-License-Identifier: Apache-2.0

"""
Test for supporting redirect caches as needed.
"""
import pytest

from tests.conftest import cache_hit, make_async_client

pytestmark = pytest.mark.asyncio


class TestPermanentRedirects(object):
    def setup(self):
        self.async_client = make_async_client()

    async def test_redirect_response_is_cached(self, url):
        await self.async_client.get(url + "permanent_redirect", allow_redirects=False)

        resp = await self.async_client.get(
            url + "permanent_redirect", allow_redirects=False
        )
        assert cache_hit(resp)

    async def test_bust_cache_on_redirect(self, url):
        await self.async_client.get(url + "permanent_redirect", allow_redirects=False)

        resp = await self.async_client.get(
            url + "permanent_redirect",
            headers={"cache-control": "no-cache"},
            allow_redirects=False,
        )
        assert not cache_hit(resp)


class TestMultipleChoicesRedirects(object):
    def setup(self):
        self.async_client = make_async_client()

    async def test_multiple_choices_is_cacheable(self, url):
        await self.async_client.get(
            url + "multiple_choices_redirect", allow_redirects=False
        )

        resp = await self.async_client.get(
            url + "multiple_choices_redirect", allow_redirects=False
        )

        assert cache_hit(resp)

    async def test_bust_cache_on_redirect(self, url):
        await self.async_client.get(
            url + "multiple_choices_redirect", allow_redirects=False
        )

        resp = await self.async_client.get(
            url + "multiple_choices_redirect",
            headers={"cache-control": "no-cache"},
            allow_redirects=False,
        )

        assert not cache_hit(resp)
